/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Dict
 *
 * The Initial Developer of the Original Code is David Costanzo
 * Portions created by the Initial Developer are Copyright (C) 2003-2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * The DictExtension.JSIRC class is originally from the JSIRC Library
 *
 * The Initial Developer of the DictExtension.JSIRC class is
 * New Dimensions Consulting, Inc.
 * Portions created by New Dimensions Consulting, Inc. are
 * Copyright (C) 1999 New Dimenstions Consulting, Inc. All
 * Rights Reserved.
 *
 * Contributor(s) to DictExtension.JSIRC class:
 *  Robert Ginda, rginda@ndcico.com, original author
 *  Peter Van der Beken, peter.vanderbeken@pandora.be, necko-only version
 *  David Costanzo
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
"use strict";

// Abstract:
//   This file is used from context.xul, message-overlay.xul and
//   dict-service-main.js, so it must only call functions that are defined
//   in those contexts.  Specifically, "alert()" is not defined when this
//   file is included from dict-service-main.js.
var EXPORTED_SYMBOLS = ['DictExtension'];

if ("undefined" == typeof(DictExtension)) {
    var DictExtension = {};
};

DictExtension.JSIRC = {
}


DictExtension.JSIRC.Connection = function()
{
    var sockServiceClass = Components.classesByID[
        "{c07e81e0-ef12-11d2-92b6-00105a1b0d64}"];

    if (!sockServiceClass)
        throw ("Couldn't get socket service class.");

    var sockService = sockServiceClass.getService();
    if (!sockService)
        throw ("Couldn't get socket service.");

    this._sockService = sockService.QueryInterface(
        Components.interfaces.nsISocketTransportService);

    this.wrappedJSObject = this;
}

DictExtension.JSIRC.Connection.prototype = {

    connect : function(host, port)
    {
        this.host = host.toLowerCase();
        this.port = port;

        this._transport = this._sockService.createTransport(null, 0, host, port, null);
        if (!this._transport)
            throw "Error creating transport.";

        var openFlags = 0;

        this._outputStream = this._transport.openOutputStream(openFlags, 4096, 32);
        if (!this._outputStream)
            throw "Error getting output stream. (no stream provider)";

        this._inputStream = this._transport.openInputStream(openFlags, 0, 0);
        if (!this._inputStream)
            throw "Error getting input stream. (no stream provider)";

        this.isConnected = true;
        return true;
    },

    disconnect : function()
    {
        if ("_inputStream" in this && this._inputStream) {
            this._inputStream.close();
        }
        if ("_outputStream" in this && this._outputStream) {
            this._outputStream.close();
        }

        this.isConnected = false;
    },

    sendData : function(str)
    {
        if (!this.isConnected)
            throw "Not Connected.";

        try {
            this._outputStream.write(str, str.length);
            return true;
        }
        catch (ex) {
            this.isConnected = false;
            throw ex;
        }
    },

    readData : function(count)
    {
        if (!this.isConnected)
            throw "Not Connected.";

        try {
            return this._scriptableInputStream.read(count);
        }
        catch (ex) {
            this.isConnected = false;
            throw ex;
        }
    },

    startAsyncRead : function(observer)
    {
        var cls = Components.classes["@mozilla.org/network/input-stream-pump;1"];
        var pump = cls.createInstance(Components.interfaces.nsIInputStreamPump);
        pump.init(this._inputStream, -1, -1, 0, 0, false);
        pump.asyncRead(new DictExtension.JSIRC.StreamListener(observer), this);
    },
};


DictExtension.JSIRC.StreamProvider = function(observer)
{
    this._observer = observer;
}

DictExtension.JSIRC.StreamProvider.prototype = {
    pendingData : "",
    isBlocked   : true,

    onDataWritable : function(request, ctxt, ostream, offset, count)
    {
        if ("isClosed" in this && this.isClosed)
            throw Components.results.NS_BASE_STREAM_CLOSED;

        if (!this.pendingData) {
            this.isBlocked = true;
            throw Components.results.NS_BASE_STREAM_WOULD_BLOCK;
        }

        var len = ostream.write (this.pendingData, this.pendingData.length);
        this.pendingData = this.pendingData.substr (len);
    },

    onStartRequest : function(request, ctxt)
    {
    },

    onStopRequest : function(request, ctxt, status)
    {
        if (this._observer)
            this._observer.onStreamClose(status);
    }
};


DictExtension.JSIRC.StreamListener = function(observer)
{
    this._observer = observer;
}

DictExtension.JSIRC.StreamListener.prototype = {

    toScriptableInputStream : function(i)
    {
        var si = Components.classes["@mozilla.org/scriptableinputstream;1"];

        si = si.createInstance();
        si = si.QueryInterface(Components.interfaces.nsIScriptableInputStream);
        si.init(i);

        return si;
    },

    onStartRequest : function(request, ctxt)
    {
    },

    onStopRequest : function(request, ctxt, status)
    {
        if (this._observer)
            this._observer.onStreamClose(status);
    },

    onDataAvailable : function(request, ctxt, inStr, sourceOffset, count)
    {
        ctxt = ctxt.wrappedJSObject;
        if (!ctxt) {
            return;
        }

        if (!("_scriptableInputStream" in ctxt)) {
            ctxt._scriptableInputStream = this.toScriptableInputStream(inStr);
        }

        if (this._observer) {
            this._observer.onStreamDataAvailable(request, inStr, sourceOffset, count);
        }
    }
};

// class constants
DictExtension.STATE = {
    Start:               0,
    WantClientAck:       1,
    WantDefinition:      2,
    WantDefinitionsDone: 3,
    WantStrategiesInfo:  4,
    WantStrategies:      5,
    WantMatchesInfo:     6,
    WantMatches:         7,
    Done:                8,
    Error:               -1
}

DictExtension.QUERYTYPE = {

    Recursive: 0, // Define word.
                  // If no definitions, match word.
                  // If only one match, define it.

    MatchOnly: 1, // Match word.

    DefineOnly: 2 // Define word.
}

DictExtension.Observer = function(listener, connection, word, database, querytype)
{
    this._listener   = listener;
    this._connection = connection;
    this._word       = word;       // encoded in UTF-8
    this._database   = database;

    this._bufferedData = '';
    this._state = DictExtension.STATE.Start;

    this._queryType = querytype;

    this._definitions  = {};

    this._createDate = new Date();
}


DictExtension.Observer.prototype = {

    // Reports that there was a protocol error.
    // This should only happen if the server doesn't obey the protocol
    // (or if there are bugs in the client state machine)
    reportStateMachineError : function ()
    {
        this._state = DictExtension.STATE.Error;

        var exception =
            'Unknown problem encountered when defining "' + DictExtension.utf8ToUcs4(this._word) + '".\n' +
            'If problem presists, please file a bug report on the Dict project.\n';

        throw exception;
    },

    // Reports that the requested database does not exist on the server
    reportUnknownDatabase : function ()
    {
        this._state = DictExtension.STATE.Error;

        var exception =
            this._connection.host + ' does not have a database called "' + this._database + '".\n'
            'Please specify a different database.\n';

        throw exception;
    },

    // Reports that the server does not have any databases installed
    reportNoDatabases : function ()
    {
        this._state = DictExtension.STATE.Error;

        var exception =
            this._connection.host + ' does not have any databases installed.\n' +
            'Please use a different server.\n';

        throw exception;
    },

    // Reports that the client received an access denied error.
    reportAccessDenied : function ()
    {
        this._state = DictExtension.STATE.Error;

        var exception =
            this._connection.host + ' requires user authentication, but ' +
            'the Dict extension does not support authentication.\n' +
            'If this feature is important to you, please let us know by filing a bug on the Dict project.';

        throw exception;
    },

    // Reports that the client received an access denied error.
    reportTemporarilyUnavailable : function ()
    {
        this._state = DictExtension.STATE.Error;

        var exception =
            this._connection.host + ' is temporarily unavailable, but expects to be available soon.\n' +
            'Please wait a short while and retry your request.';

        throw exception;
    },


    // Reports that the client received an access denied error.
    reportServerShuttingDown : function ()
    {
        this._state = DictExtension.STATE.Error;

        var exception =
            this._connection.host + ' is being shut down and will not be able to define\n' +
            'words in the near-future.\n' +
            'Please change your preferences to use a different server.';

        throw exception;
    },

    // Pops a DICT response line from the _bufferedInput member variable.
    // On success, returns true and sets this._responseLine, this._responseCode,
    // and this._responseLineBody.
    // Return false, if there wasn't a complete response line.
    popResponseLine : function ()
    {
        var responseLinePattern = /^(\d+)\s*(.*)\n/;

        var result = this._bufferedData.match(responseLinePattern);
        if (result == null) {
            return false;
        }

        this._responseLine     = result[0];
        this._responseCode     = result[1];
        this._responseLineBody = result[2];

        // remove the response line from the buffered input
        var lineEndIndex   = this._bufferedData.indexOf('\n');
        this._bufferedData = this._bufferedData.slice(lineEndIndex + 1);

        // Handle error codes that should result in immediately
        // reporting an error to the user regardless of the current state.
        if (this._responseCode == 530 ||
            this._responseCode == 531 ||
            this._responseCode == 532) {
            // access denied
            this.reportAccessDenied();
        }
        else if (this._responseCode == 420) {
            // server is temporarily unavailable
            this.reportTemporarilyUnavailable();
        }
        else if (this._responseCode == 421) {
            // server is temporarily unavailable
            this.reportServerShuttingDown();
        }
        else if (this._responseCode == 550) {
            // invalid database
            this.reportUnknownDatabase();
        }
        else if (this._responseCode == 554) {
            // no databases present
            this.reportNoDatabases();
        }

        return true;
    },

    // Performs a graceful shutdown
    gracefulShutdown : function ()
    {
        this._connection.sendData('QUIT\r\n');
        this._bufferedData = '';
        this._state = DictExtension.STATE.Done;

        this._connection.disconnect();
    },

    // Returns a word in a format that is suitable for inserting into
    // DICT protocol requests.
    quoteWord : function(utf8Word) {
        return '"' + utf8Word.replace(/"/g, '\\"') + '"'; /*"*/
    },

    // Unquotes a string response from a DICT server.
    unquoteWord : function(utf8Word) {
        // words can be single quoted or double-quoted
        utf8Word = utf8Word.replace(/^["'](.*)['"]$/, '$1');

        // characters can be escaped.
        utf8Word = utf8Word.replace(/\\(.)/g, '$1');

        // Work-around problems where the server sends malformed UTF-8.
        utf8Word = DictExtension.ucs4ToUtf8(DictExtension.utf8ToUcs4(utf8Word));

        return utf8Word;
    },


    // Parses and returns all buffered input up to (but not including)
    // the sequence CRLF . CRLF.  The multiline input is then removed
    // from the buffered input.
    // Returns null if the multiline input is not complete.
    //
    // The DICT protocol uses multiline input of this form to return
    // strategies, matches, and definitions.
    popMultilineInput : function ()
    {
        // NOTE: We should look for \r\n.\r\n, but all "\r" are removed
        //       from the input stream to make regexp parsing easier.
        const sentinal = "\n.\n";

        // Check to see if we have the entire list
        var endIndex = this._bufferedData.indexOf(sentinal);
        if (endIndex == -1) {
            return null;  // the multiline input isn't complete
        }

        // pop off and return the multiline input
        var multilineData = this._bufferedData.slice(0, endIndex);

        // remove the multiline input and sentinal sequence from the buffered input
        this._bufferedData = this._bufferedData.slice(endIndex + sentinal.length);

        // According to RFC 2229, Section 2.4.3, if a line of text begins
        // with two periods, the client must treat this as a single,
        // literal period.
        multilineData = multilineData.replace(/^\.\./gm, '.');

        return multilineData;
    },

    onStreamDataAvailable : function (
        request,
        inStream,
        offset,
        count
    )
    {
        if (this._state == DictExtension.STATE.Done ||
            this._state == DictExtension.STATE.Error) {
            // sometimes if we perform a graceful shutdown, this handler
            // gets called again.  Any operation will throw an exception,
            // so it's best to just ignore the call.
            return;
        }

        try {
            // read the data from the socket into our buffer
            var newData = this._connection.readData(count);

            // strip out carriage returns, since they complicate regexp usage
            this._bufferedData += newData.replace(/\r/g, '');

            // do {
            //   {handle data}
            // } while( {there's enough data to handle} );
            do {
                var previousLength = this._bufferedData.length;

                var matchExp;
                var word;

                switch (this._state) {

                case DictExtension.STATE.Start:
                    // get the 220 "hello" response that is sent when we connect

                    if (this.popResponseLine()) {

                        if (this._responseCode == 220) {
                            // 220 <text> <capabilities> <msg-id> CRLF
                            // (this is the session init sequence)
                            this._state = DictExtension.STATE.WantClientAck;
                        }
                        else {
                            this.reportStateMachineError();
                        }
                    }
                    break;


                case DictExtension.STATE.WantClientAck:
                    // get the 250 ok resonse for our "CLIENT" command

                    if (this.popResponseLine()) {
                        if (this._responseCode == 250) {

                            if (this._queryType == DictExtension.QUERYTYPE.MatchOnly) {
                                // We have pipelined a MATCH request.
                                // We expect to see the matches next.
                                this._state = DictExtension.STATE.WantMatchesInfo;
                            }
                            else {
                                // We have pipelined a DEFINE request.
                                // We expect to see the definitions next.
                                this._state = DictExtension.STATE.WantDefinitionsDone;
                            }
                        }
                        else {
                            this.reportStateMachineError();
                        }
                    }
                    break;

                case DictExtension.STATE.WantDefinitionsDone:

                    if (this.popResponseLine()) {
                        if (this._responseCode == 150) {
                            // 150: n definitions returned
                        }
                        else if (this._responseCode == 151) {
                            // 151: <word> <database> <name>
                            // The definition follows this line
                            this._state = DictExtension.STATE.WantDefinition;
                        }
                        else if (this._responseCode == 250) {

                            // We have all definitions
                            this._listener.onHaveDefinitions(
                                this._word,
                                this._definitions);

                            this.gracefulShutdown();
                        }
                        else if (this._responseCode == 550) {
                            // 550 invalid database

                            // TODO: Handle this gracefully
                            this.reportStateMachineError();
                        }
                        else if (this._responseCode == 552) {
                            // 552 no match

                            // Present the user with a nearest match

                            // The "." strategy means "use whatever is best for
                            // for interactive spell-checking", according to
                            // RFC 2229, section 3.3.1.
                            //
                            this._connection.sendData('MATCH ' + this._database + ' . ' + this.quoteWord(this._word) + '\r\n');
                            this._state = DictExtension.STATE.WantMatchesInfo;
                        }
                        else {
                            this.reportStateMachineError();
                        }
                    }
                    break;

                case DictExtension.STATE.WantDefinition:

                    // 151 <word> <database> <full database name> CRLF
                    // <definition line 1>   CRLF
                    // <definition line 2>   CRLF
                    // . . .
                    // <definition line n>   CRLF
                    // . CRLF

                    // determine if we have the entire definition
                    var definitionText = this.popMultilineInput();
                    if (definitionText == null) {
                        break; // wait for the rest of the definition
                    }

                    var definition = {};

                    // 151 <word> <database> <full database name> CRLF
                    //
                    // Word can be:
                    //     non-whitespace
                    //     " (non-double quote) "
                    //     ' (non-quote) '
                    //
                    var result = this._responseLineBody.match(/^([^'"\s]\S*|"[^"]*"|'[^']*')\s+\S+\s+([^'"\s]\S*|"[^"]*"|'[^']*').*$/);
                    if (result != null) {
                        definition.word   = this.unquoteWord(result[1]);
                        definition.source = this.unquoteWord(result[2]);
                    }

                    // Store the definition
                    definition.text = definitionText;

                    // Add this definition to the set of definitions.
                    // this._definitions maps the source name to an array of
                    // definition objects.
                    if (this._definitions[definition.source] != null) {
                        this._definitions[definition.source].push(definition);
                    }
                    else {
                        this._definitions[definition.source] = new Array ();
                        this._definitions[definition.source].push(definition);
                    }

                    this._state = DictExtension.STATE.WantDefinitionsDone;
                    break;

                case DictExtension.STATE.WantStrategiesInfo:

                    // DEAD CODE:
                    //   This state is not reachable, since the code never
                    //   sends the "SHOW STRAT" command.  However,
                    //   this state code is correct, so I have kept it intact.

                    if (this.popResponseLine()) {
                        if (this._responseCode == 111) {
                            // 111 n stragegies available
                            this._state = DictExtension.STATE.WantStrategies;
                        }
                        else if (this._responseCode == 555) {
                            // 555 No strategies available
                            this.gracefulShutdown();
                            this._listener.onNoDefinitionsFound(this._word);
                        }
                        else {
                            // unexpected response
                            this.reportStateMachineError();
                        }
                    }
                    break;

                case DictExtension.STATE.WantStrategies:

                    // DEAD CODE:
                    //   This state is not reachable, since the code never
                    //   sends the "SHOW STRAT" command.  However,
                    //   the state code is correct, so I have kept it intact.

                    // <strategy 1> <description 1> CRLF
                    // <strategy 2> <description 2> CRLF
                    // . . .
                    // <strategy n> <description n> CRLF
                    // . CRLF

                    // determine if we have the entire strategy list
                    var strategies = this.popMultilineInput();
                    if (strategies == null) {
                        break; // wait for the rest of the strategies
                    }


                    // search for the "Levenshtein distance one"
                    var haveLevStrategy = false;
                    // <strategy> <description> CRLF
                    matchExp = /^(\S+)\s+(.*)$/gm;
                    while ((result = matchExp.exec(strategies)) != null) {
                        var strategy    = result[1];
                        var description = result[2];

                        if (strategy == 'lev') {
                            haveLevStrategy = true;
                        }
                    }

                    if (haveLevStrategy) {
                        // Use the "lev" strategy to find nearest matches
                        this._connection.sendData('MATCH ' + this._database + ' lev ' + this.quoteWord(this._word) + '\r\n');
                        this._state = DictExtension.STATE.WantMatchesInfo;
                    }
                    else {
                        // No good matching strategies
                        this.gracefulShutdown();

                        this._listener.onNoDefinitionsFound(this._word);
                    }

                    break;

                case DictExtension.STATE.WantMatchesInfo:

                    if (this.popResponseLine()) {
                        if (this._responseCode == 250) {
                            // 250 ok
                            // This is from the MATCH request.  Ignore it.
                        }
                        else if (this._responseCode == 152) {
                            // 152 n matches found
                            this._state = DictExtension.STATE.WantMatches;
                        }
                        else if (this._responseCode == 552) {
                            // 552 no match
                            this.gracefulShutdown();
                            if (this._queryType == DictExtension.QUERYTYPE.MatchOnly) {
                                this._listener.onNoMatchesFound(this._word);
                            }
                            else {
                                this._listener.onNoDefinitionsFound(this._word);
                            }
                        }
                        else {
                            // unexpected response
                            this.reportStateMachineError();
                        }
                    }
                    break;

                case DictExtension.STATE.WantMatches:

                    // <database 1> <word 1> CRLF
                    // <database 2> <word 2> CRLF
                    // . . .
                    // <database n> <word n> CRLF
                    // . CRLF

                    // determine if we have the entire list of matches
                    var matchesResponse = this.popMultilineInput();
                    if (matchesResponse == null) {
                        break; // wait for the rest of the matches
                    }

                    // <database> <word> CRLF
                    var distinctMatches = {};
                    matchExp = /^(\S+)\s+(.*)$/gm;
                    while ((result = matchExp.exec(matchesResponse)) != null) {
                        var database = this.unquoteWord(result[1]);
                        var match    = this.unquoteWord(result[2]);

                        // Normalize the match so that we can keep
                        // track of distinct matches.  For example,
                        // "Australia" and "australia" should be
                        // considered the same word.
                        //
                        // To normalize the word, we need to convert
                        // back into UCS-4 so that toLocaleUpperCase()
                        // will work.
                        var ucs4match = DictExtension.utf8ToUcs4(match);
                        distinctMatches[ucs4match.toLocaleUpperCase()] = match;
                    }

                    // Repackage the distinct matches dictionary into a
                    // form which the caller expects--a dictionary whose
                    // keys are UTF-8 words.
                    var matches      = {};
                    var totalMatches = 0;
                    for (var normalizedWord in distinctMatches) {

                        matches[distinctMatches[normalizedWord]] = true;
                        totalMatches++;
                    }

                    if (this._queryType == DictExtension.QUERYTYPE.MatchOnly) {
                        // We are only looking for matches.
                        if (totalMatches == 0) {
                            // This is probably a bug in either the server or the
                            // client, since we should not have transitioned to
                            // the WantMatches state if there were no matches.
                            this._listener.onNoMatchesFound(this._word);
                            this.gracefulShutdown();
                        }
                        else {
                            // We found some matches.
                            this._listener.onHaveClosestMatches(this._word, matches);
                            this.gracefulShutdown();
                        }
                    }
                    else {
                        // this is a "recursive" query.

                        if (totalMatches == 0) {
                            // There are no matches.

                            // This is probably a bug in either the server or the
                            // client, since we should not have transitioned to
                            // the WantMatches state if there were no matches.
                            this.gracefulShutdown();

                            this._listener.onNoDefinitionsFound(this._word);
                        }
                        else if (totalMatches == 1) {
                            // There is exactly one match.
                            // Rather than present just one match, we can save the
                            // user some effort and just request the definition
                            // of the match.

                            for (word in matches) {
                                // Request a definition for the only match
                                this._connection.sendData('DEFINE ' + this._database + ' ' + this.quoteWord(word) + '\r\n');
                                this._word  = word;
                                this._state = DictExtension.STATE.WantClientAck;
                            }
                        }
                        else {
                            // There is more than one close match
                            this._listener.onHaveClosestMatches(this._word, matches);

                            this.gracefulShutdown();
                        }
                    }

                    break;

                case DictExtension.STATE.Done:
                    break;

                default:
                    reportStateMachineError();
                    break;
                }

            } while (previousLength != this._bufferedData.length);
        }
        catch (ex) {
            this._connection.disconnect();
            this._listener.onError(ex.toString());
        }
    },

    onStreamClose : function (
        status
    )
    {
        var closeDate = new Date();
        var delta = closeDate.getTime() - this._createDate.getTime();
        //alert('query took ' + delta + ' ms');

        if (this._state != DictExtension.STATE.Done &&
            this._state != DictExtension.STATE.Error) {
            this._listener.onError(
                'Unable to connect to "' + this._connection.host + '".\n' +
                'If the problem persists, try using a different server.');
        }
    },
};


// This file contains helper functions for converting from UTF-8 to UCS-4
// and vice-versa.
//
//   UCS-4 range (hex.)    UTF-8 octet sequence (binary)
//   0000 0000-0000 007F   0xxxxxxx
//   0000 0080-0000 07FF   110xxxxx 10xxxxxx
//   0000 0800-0000 FFFF   1110xxxx 10xxxxxx 10xxxxxx
//   0001 0000-001F FFFF   11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
//   0020 0000-03FF FFFF   111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
//   0400 0000-7FFF FFFF   1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
//

// utf8ToUcs4() converts an octet stream that is
// encoded in UTF-8 into a Unicode encoding of the same
// stream.
DictExtension.utf8ToUcs4 = function(utf8String)
{
    // Before we do any conversion, let's see if there are
    // any character codes outside of the ASCII-7 range.
    // If not, then the string doesn't need to be converted.
    if (!utf8String.match(/[\x80-\xFF]/)) {
        return utf8String;
    }

    // The worker function that converts from UTF-8 to UCS-4.
    // It has no optimizations and does a complete conversion on every character.
    var convertUtf8ToUcs4Worker = function(utf8String) {

        var encodedString = '';
        var octetsRemainingInCharacterCode = 0;
        var characterCode = 0;

        for (var i = 0; i < utf8String.length; i++) {

            var inputCharCode = utf8String.charCodeAt(i);

            if ((inputCharCode & 0x80) == 0x00) {
                // this can be encoded with ASCII-7
                encodedString += utf8String[i];

                // reset the multi-octet stream accumulator.
                // if it was not zero, then the input was bad.
                octetsRemainingInCharacterCode = 0;
            }
            else if ((inputCharCode & 0xC0) == 0x80) {
                // 10xxxxxx

                // this is a continuation of a multi-octet sequence
                characterCode = (characterCode << 6) | (inputCharCode & 0x3F);

                octetsRemainingInCharacterCode--;
                if (octetsRemainingInCharacterCode == 0) {
                    // This is the final octet in the mutli-octet code.
                    encodedString += String.fromCharCode(characterCode);
                }
                else if (octetsRemainingInCharacterCode < 0) {
                    // bad input
                    octetsRemainingInCharacterCode = 0;
                }
            }
            else if ((inputCharCode & 0xE0) == 0xC0) {
                // 110xxxxx 10xxxxxx
                characterCode = inputCharCode & 0x1F;
                octetsRemainingInCharacterCode = 1;
            }
            else if ((inputCharCode & 0xF0) == 0xE0) {
                // 1110xxxx 10xxxxxx 10xxxxxx
                characterCode = inputCharCode & 0x0F;
                octetsRemainingInCharacterCode = 2;
            }
            else if ((inputCharCode & 0xF8) == 0xF0) {
                // 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
                characterCode = inputCharCode & 0x07;
                octetsRemainingInCharacterCode = 3;
            }
            else if ((inputCharCode & 0xFC) == 0xF8) {
                // 111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
                characterCode = inputCharCode & 0x03;
                octetsRemainingInCharacterCode = 4;
            }
            else if ((inputCharCode & 0xFE) == 0xFC) {
                // 1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
                characterCode = inputCharCode & 0x01;
                octetsRemainingInCharacterCode = 5;
            }
            else {
                // Illegal UTF-8 code.  Drop it.
            }
        }

        if (octetsRemainingInCharacterCode != 0) {

            // The UTF-8 seems to have been cut short.
            // kassiopeia.juls.savba.sk sometimes does this.
            // Rather than reject the server, we are tolerant of bad input.
            encodedString += String.fromCharCode(characterCode);
        }

        return encodedString;
    }

    // Note that we could just call convertUtf8ToUcs4Worker()
    // on the entire string, but this routine is very slow.
    // Instead, it is faster to use a regular expression
    // to separate the octet-sequences that must be replaced
    // while preserving long runs of characters that don't need conversion.

    // This makes the routine run in O(n^2) of the characters that
    // must be replaced, instead of O(n^2) of all characters.
    // When the text is mostly ASCII-7, this is a significant savings.

    var encodedString = '';

    var result;
    var matchExp = /([\x00-\x7F]*)([\x80-\xFF]+)([\x00-\x7F]*)/g;
    while ((result = matchExp.exec(utf8String)) != null) {

        var singleOctetCodesBefore = result[1];
        var multiOctetCodes        = result[2];
        var singleOctetCodesAfter  = result[3];

        encodedString +=
            singleOctetCodesBefore +
            convertUtf8ToUcs4Worker(multiOctetCodes) +
            singleOctetCodesAfter;
    }

    return encodedString;
}


// returns a UTF-8 representation of ucs4String
DictExtension.ucs4ToUtf8 = function(ucs4String)
{
    var utf8String = '';
    for (var i = 0; i < ucs4String.length; i++) {
        var charCode = ucs4String.charCodeAt(i);

        if (0x7FFFFFFF < charCode) {
            // This byte cannot be encoded in UTF-8.
            // Skip it
            // NOTE: this should never happen in practice
            // because JavaScript uses UCS-4.
        }
        else if (0x3FFFFFF < charCode) {
            // six-byte UTF-8 sequence
            // 1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
            utf8String += String.fromCharCode(
                0xFC | ((charCode >> 30) & 0x01),
                0x80 | ((charCode >> 24) & 0x3F),
                0x80 | ((charCode >> 18) & 0x3F),
                0x80 | ((charCode >> 12) & 0x3F),
                0x80 | ((charCode >>  6) & 0x3F),
                0x80 | ((charCode >>  0) & 0x3F));
        }
        else if (0x1FFFFF < charCode) {
            // five-byte UTF-8 sequence
            // 111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
            utf8String += String.fromCharCode(
                0xF8 | ((charCode >> 24) & 0x03),
                0x80 | ((charCode >> 18) & 0x3F),
                0x80 | ((charCode >> 12) & 0x3F),
                0x80 | ((charCode >>  6) & 0x3F),
                0x80 | ((charCode >>  0) & 0x3F));
        }
        else if (0xFFFF < charCode) {
            // four-byte UTF-8 sequence
            // 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
            utf8String += String.fromCharCode(
                0xF0 | ((charCode >> 18) & 0x07),
                0x80 | ((charCode >> 12) & 0x3F),
                0x80 | ((charCode >>  6) & 0x3F),
                0x80 | ((charCode >>  0) & 0x3F));
        }
        else if (0x7FF < charCode) {
            // three-byte UTF-8 sequence
            // 1110xxxx 10xxxxxx 10xxxxxx
            utf8String += String.fromCharCode(
                0xE0 | ((charCode >> 12) & 0x0F),
                0x80 | ((charCode >>  6) & 0x3F),
                0x80 | ((charCode >>  0) & 0x3F));
        }
        else if (0x7F < charCode) {
            // two-byte UTF-8 sequence
            // 110xxxxx 10xxxxxx
            utf8String += String.fromCharCode(
                0xC0 | ((charCode >>  6) & 0x1F),
                0x80 | ((charCode >>  0) & 0x3F));
        }
        else {
            // one-byte UTF-8 sequence
            utf8String += ucs4String[i];
        }
    }

    return utf8String;
}

DictExtension.Client = function()
{
}

DictExtension.Client.prototype =
{
    // defineWord() is the main entry point to the DICT client.
    // This function contacts the DICT server and asks for the
    // definition of a given word.  It is asynchornous and the
    // definitions (or errors) are returned by callbacks on
    // the "listener" parameter.
    //
    // host -
    //     The host running the DICT server
    //
    // port -
    //     The port on which the DICT server is listening.
    //     This is usually 2628.
    //
    // word -
    //     The word to define.  This should be encoded in USC-4.
    //
    // listener -
    //     A special "listener" object which handles the server's response.
    //     This object must define the following functions:
    //
    //       onHaveDefinitions: function (word, definitions)
    //          Called when DictExtension.Client has the definitions.
    //
    //          word -
    //              The word that was actually defined, encoded in UTF-8.
    //              This may be different from the requested word, if
    //              the exact word was not found and only one match
    //              existed.
    //
    //          definitions -
    //              An associated array that maps the name of the definition's source
    //              (the dictionary) to an array of definitions.
    //              Each definition is an object with one field "text",
    //              which is a UTF-8 encoded definition of the word.
    //
    //              The following code shows how to iterate over all definitions:
    //
    //                for (var source in definitions) {
    //
    //                  var i;
    //                  for (i = 0; i < definitions[source].length; i++) {
    //
    //                    var def = definitions[source][i].text;
    //                    <do something with def>
    //                  }
    //                }
    //
    //       onHaveClosestMatches: function (word, matches)
    //          Called when DictExtension.Client has a list of close matches.
    //
    //          word    -
    //              The word for which there are close matches.
    //              This is encoded in UTF-8.
    //
    //          matches -
    //              An associative array whose keys are strings,
    //              each of which is a close match for a word
    //              that has a definition.
    //              The keys are encoded in UTF-8.
    //
    //       onNoDefinitionsFound : function (word)
    //          Called when the DictExtension.Client can not find any definitions.
    //
    //          word -
    //              the word passed into defineWord() encoded in UTF-8.
    //
    //       onError : function (error)
    //          Called when the DictExtension.Client encountered an unexpected error.
    //
    //          error -
    //              A plain-text explaination of what went wrong.
    //              This string is encoded in USC-4.
    //
    // database - [optional]
    //    The database to define the word.
    //    If this parameter is not given, '*' will be used.
    //
    defineWord : function(host, port, word, listener, database)
    {
        // check to see if we are in an offline mode
        var ioService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);

        if (ioService.offline) {
            listener.onError(
                'You cannot define words while offline.\n' +
                'To go online, uncheck "Work Offline" from the File menu.');
            return;
        }


        // if no database was specified, default to '*' (all databases)
        if (database == null) {
            database = '*';
        }

        var connection = new DictExtension.JSIRC.Connection();

        connection.connect(host, port);

        // construct a UTF-8 representation of word
        var utf8Word = DictExtension.ucs4ToUtf8(word);

        var observer = new DictExtension.Observer(
            listener,
            connection,
            utf8Word,
            database,
            DictExtension.QUERYTYPE.Recursive);

        try {
            // To save time, pipe-line the CLIENT and DEFINE commands,
            // rather than wait for the "220 ok" handshake.
            // The server is required to support command pipe-lining.
            // This makes a measurable speed gain when there is high latency.
            connection.sendData('CLIENT Mozilla Dict extension\r\n');
            connection.sendData('DEFINE ' + database + ' ' + observer.quoteWord(utf8Word) + '\r\n');
            connection.startAsyncRead(observer);
        }
        catch (ex) {
            listener.onError(
                'Unable to connect to "' + host + '".\n' +
                'If the problem persists, try using a different server.');
        }
    },

    // matchWord() is a secondary entry point to the DICT client.
    // This function contacts the DICT server and asks for the
    // closest matches of a given word.  It is asynchornous and the
    // matches (or errors) are returned by callbacks on the "listener"
    // parameter.
    //
    // host -
    //     The host running the DICT server
    //
    // port -
    //     The port on which the DICT server is listening.
    //     This is usually 2628.
    //
    // word -
    //     The word to find matches for.  This should be encoded in USC-4.
    //
    // listener -
    //     A special "listener" object which handles the server's response.
    //     This object must define the following functions:
    //
    //       onHaveClosestMatches: function (word, matches)
    //          Called when DictExtension.Client has a list of close matches.
    //
    //          word    -
    //              The word for which there are close matches.
    //              This is encoded in UTF-8.
    //
    //          matches -
    //              An associative array whose keys are strings,
    //              each of which is a close match for a word
    //              that has a definition.
    //              The keys are encoded in UTF-8.
    //
    //       onNoMatchesFound : function (word)
    //          Called when the DictExtension.Client can not find any matches.
    //
    //          word -
    //              the word passed into defineWord() encoded in UTF-8.
    //
    //       onError : function (error)
    //          Called when the DictExtension.Client encountered an unexpected error.
    //
    //          error -
    //              A plain-text explaination of what went wrong.
    //              This string is encoded in USC-4.
    //
    // database - [optional]
    //    The database to define the word.
    //    If this parameter is not given, all databases on host will be
    //    searched.
    //
    // strategy - [optional]
    //    The strategy to use when matching.
    //    If this parameter is not given, a strategy that works well
    //    for catching simple misspellings will be used.
    //
    matchWord : function(host, port, word, listener, database, strategy)
    {

        // check to see if we are in an offline mode
        var ioService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);

        if (ioService.offline) {
            listener.onError(
                'You cannot define words while offline.\n' +
                'To go online, uncheck "Work Offline" from the File menu.');
            return;
        }


        // if no database was specified, default to '*' (all databases)
        if (database == null) {
            database = '*';
        }

        // if no strategy was specified, default to '.'
        if (strategy == null) {
            strategy = '.';
        }

        var connection = new DictExtension.JSIRC.Connection();

        connection.connect(host, port);

        // construct a UTF-8 representation of word
        var utf8Word = DictExtension.ucs4ToUtf8(word);

        var observer = new DictExtension.Observer(
            listener,
            connection,
            utf8Word,
            database,
            DictExtension.QUERYTYPE.MatchOnly);

        try {
            // To save time, pipe-line the CLIENT and MATCH commands,
            // rather than wait for the "220 ok" handshake.
            // The server is required to support command pipe-lining.
            // This makes a measurable speed gain when there is high latency.
            connection.sendData('CLIENT Mozilla Dict extension\r\n');
            connection.sendData('MATCH ' + database + ' ' + strategy + ' ' + observer.quoteWord(utf8Word) + '\r\n');

            connection.startAsyncRead(observer);
        }
        catch (ex) {
            listener.onError(
                'Unable to connect to "' + host + '".\n' +
                'If the problem persists, try using a different server.');
        }
    }
}
